=== modified file 'TODO'
--- TODO	2019-02-13 08:45:09 +0000
+++ TODO	2019-03-17 16:48:56 +0000
@@ -86,7 +86,6 @@
 *** TODO [#C] In Python 3.3, use shlex.quote() instead of re.escape()
 
 * mandos-ctl
-*** Handle "no D-Bus server" and/or "no Mandos server found" better
 ** TODO Remove old string_to_delta format				  :2:
 
 * TODO mandos-dispatch

=== modified file 'mandos-ctl'
--- mandos-ctl	2019-03-09 16:14:06 +0000
+++ mandos-ctl	2019-03-17 16:48:56 +0000
@@ -64,11 +64,12 @@
 
 locale.setlocale(locale.LC_ALL, "")
 
-domain = "se.recompile"
-busname = domain + ".Mandos"
-server_path = "/"
-server_interface = domain + ".Mandos"
-client_interface = domain + ".Mandos.Client"
+dbus_busname_domain = "se.recompile"
+dbus_busname = dbus_busname_domain + ".Mandos"
+server_dbus_path = "/"
+server_dbus_interface = dbus_busname_domain + ".Mandos"
+client_dbus_interface = dbus_busname_domain + ".Mandos.Client"
+del dbus_busname_domain
 version = "1.8.3"
 
 
@@ -78,13 +79,134 @@
     dbus.OBJECT_MANAGER_IFACE = "org.freedesktop.DBus.ObjectManager"
 
 
-def milliseconds_to_string(ms):
-    td = datetime.timedelta(0, 0, 0, ms)
-    return ("{days}{hours:02}:{minutes:02}:{seconds:02}"
-            .format(days="{}T".format(td.days) if td.days else "",
-                    hours=td.seconds // 3600,
-                    minutes=(td.seconds % 3600) // 60,
-                    seconds=td.seconds % 60))
+def main():
+    parser = argparse.ArgumentParser()
+
+    add_command_line_options(parser)
+
+    options = parser.parse_args()
+
+    check_option_syntax(parser, options)
+
+    clientnames = options.client
+
+    if options.debug:
+        log.setLevel(logging.DEBUG)
+
+    bus = dbus.SystemBus()
+
+    mandos_dbus_object = get_mandos_dbus_object(bus)
+
+    mandos_serv = dbus.Interface(
+        mandos_dbus_object, dbus_interface=server_dbus_interface)
+    mandos_serv_object_manager = dbus.Interface(
+        mandos_dbus_object, dbus_interface=dbus.OBJECT_MANAGER_IFACE)
+
+    managed_objects = get_managed_objects(mandos_serv_object_manager)
+
+    all_clients = {}
+    for path, ifs_and_props in managed_objects.items():
+        try:
+            all_clients[path] = ifs_and_props[client_dbus_interface]
+        except KeyError:
+            pass
+
+    # Compile dict of (clientpath: properties) to process
+    if not clientnames:
+        clients = all_clients
+    else:
+        clients = {}
+        for name in clientnames:
+            for objpath, properties in all_clients.items():
+                if properties["Name"] == name:
+                    clients[objpath] = properties
+                    break
+            else:
+                log.critical("Client not found on server: %r", name)
+                sys.exit(1)
+
+    commands = commands_from_options(options)
+
+    for command in commands:
+        command.run(clients, bus, mandos_serv)
+
+
+def add_command_line_options(parser):
+    parser.add_argument("--version", action="version",
+                        version="%(prog)s {}".format(version),
+                        help="show version number and exit")
+    parser.add_argument("-a", "--all", action="store_true",
+                        help="Select all clients")
+    parser.add_argument("-v", "--verbose", action="store_true",
+                        help="Print all fields")
+    parser.add_argument("-j", "--dump-json", action="store_true",
+                        help="Dump client data in JSON format")
+    enable_disable = parser.add_mutually_exclusive_group()
+    enable_disable.add_argument("-e", "--enable", action="store_true",
+                                help="Enable client")
+    enable_disable.add_argument("-d", "--disable",
+                                action="store_true",
+                                help="disable client")
+    parser.add_argument("-b", "--bump-timeout", action="store_true",
+                        help="Bump timeout for client")
+    start_stop_checker = parser.add_mutually_exclusive_group()
+    start_stop_checker.add_argument("--start-checker",
+                                    action="store_true",
+                                    help="Start checker for client")
+    start_stop_checker.add_argument("--stop-checker",
+                                    action="store_true",
+                                    help="Stop checker for client")
+    parser.add_argument("-V", "--is-enabled", action="store_true",
+                        help="Check if client is enabled")
+    parser.add_argument("-r", "--remove", action="store_true",
+                        help="Remove client")
+    parser.add_argument("-c", "--checker",
+                        help="Set checker command for client")
+    parser.add_argument("-t", "--timeout", type=string_to_delta,
+                        help="Set timeout for client")
+    parser.add_argument("--extended-timeout", type=string_to_delta,
+                        help="Set extended timeout for client")
+    parser.add_argument("-i", "--interval", type=string_to_delta,
+                        help="Set checker interval for client")
+    approve_deny_default = parser.add_mutually_exclusive_group()
+    approve_deny_default.add_argument(
+        "--approve-by-default", action="store_true",
+        default=None, dest="approved_by_default",
+        help="Set client to be approved by default")
+    approve_deny_default.add_argument(
+        "--deny-by-default", action="store_false",
+        dest="approved_by_default",
+        help="Set client to be denied by default")
+    parser.add_argument("--approval-delay", type=string_to_delta,
+                        help="Set delay before client approve/deny")
+    parser.add_argument("--approval-duration", type=string_to_delta,
+                        help="Set duration of one client approval")
+    parser.add_argument("-H", "--host", help="Set host for client")
+    parser.add_argument("-s", "--secret",
+                        type=argparse.FileType(mode="rb"),
+                        help="Set password blob (file) for client")
+    approve_deny = parser.add_mutually_exclusive_group()
+    approve_deny.add_argument(
+        "-A", "--approve", action="store_true",
+        help="Approve any current client request")
+    approve_deny.add_argument("-D", "--deny", action="store_true",
+                              help="Deny any current client request")
+    parser.add_argument("--debug", action="store_true",
+                        help="Debug mode (show D-Bus commands)")
+    parser.add_argument("--check", action="store_true",
+                        help="Run self-test")
+    parser.add_argument("client", nargs="*", help="Client name")
+
+
+def string_to_delta(interval):
+    """Parse a string and return a datetime.timedelta"""
+
+    try:
+        return rfc3339_duration_to_delta(interval)
+    except ValueError as e:
+        log.warning("%s - Parsing as pre-1.6.1 interval instead",
+                    ' '.join(e.args))
+    return parse_pre_1_6_1_interval(interval)
 
 
 def rfc3339_duration_to_delta(duration):
@@ -110,31 +232,31 @@
     >>> rfc3339_duration_to_delta("")
     Traceback (most recent call last):
     ...
-    ValueError: Invalid RFC 3339 duration: u''
+    ValueError: Invalid RFC 3339 duration: ""
     >>> # Must start with "P":
     >>> rfc3339_duration_to_delta("1D")
     Traceback (most recent call last):
     ...
-    ValueError: Invalid RFC 3339 duration: u'1D'
+    ValueError: Invalid RFC 3339 duration: "1D"
     >>> # Must use correct order
     >>> rfc3339_duration_to_delta("PT1S2M")
     Traceback (most recent call last):
     ...
-    ValueError: Invalid RFC 3339 duration: u'PT1S2M'
+    ValueError: Invalid RFC 3339 duration: "PT1S2M"
     >>> # Time needs time marker
     >>> rfc3339_duration_to_delta("P1H2S")
     Traceback (most recent call last):
     ...
-    ValueError: Invalid RFC 3339 duration: u'P1H2S'
+    ValueError: Invalid RFC 3339 duration: "P1H2S"
     >>> # Weeks can not be combined with anything else
     >>> rfc3339_duration_to_delta("P1D2W")
     Traceback (most recent call last):
     ...
-    ValueError: Invalid RFC 3339 duration: u'P1D2W'
+    ValueError: Invalid RFC 3339 duration: "P1D2W"
     >>> rfc3339_duration_to_delta("P2W2H")
     Traceback (most recent call last):
     ...
-    ValueError: Invalid RFC 3339 duration: u'P2W2H'
+    ValueError: Invalid RFC 3339 duration: "P2W2H"
     """
 
     # Parsing an RFC 3339 duration with regular expressions is not
@@ -211,23 +333,12 @@
                 break
         else:
             # No currently valid tokens were found
-            raise ValueError("Invalid RFC 3339 duration: {!r}"
+            raise ValueError("Invalid RFC 3339 duration: \"{}\""
                              .format(duration))
     # End token found
     return value
 
 
-def string_to_delta(interval):
-    """Parse a string and return a datetime.timedelta"""
-
-    try:
-        return rfc3339_duration_to_delta(interval)
-    except ValueError as e:
-        log.warning("%s - Parsing as pre-1.6.1 interval instead",
-                    ' '.join(e.args))
-    return parse_pre_1_6_1_interval(interval)
-
-
 def parse_pre_1_6_1_interval(interval):
     """Parse an interval string as documented by Mandos before 1.6.1,
     and return a datetime.timedelta
@@ -271,400 +382,6 @@
     return value
 
 
-## Classes for commands.
-
-# Abstract classes first
-class Command(object):
-    """Abstract class for commands"""
-    def run(self, mandos, mandos_objmgr, clients):
-        """Normal commands should implement run_on_one_client(), but
-        commands which want to operate on all clients at the same time
-        can override this run() method instead."""
-        self.mandos = mandos
-        for client, properties in clients.items():
-            self.run_on_one_client(client, properties)
-
-class PrintCmd(Command):
-    """Abstract class for commands printing client details"""
-    all_keywords = ("Name", "Enabled", "Timeout", "LastCheckedOK",
-                    "Created", "Interval", "Host", "KeyID",
-                    "Fingerprint", "CheckerRunning", "LastEnabled",
-                    "ApprovalPending", "ApprovedByDefault",
-                    "LastApprovalRequest", "ApprovalDelay",
-                    "ApprovalDuration", "Checker", "ExtendedTimeout",
-                    "Expires", "LastCheckerStatus")
-    def run(self, mandos, mandos_objmgr, clients):
-        print(self.output(mandos_objmgr, clients.values()))
-
-class PropertyCmd(Command):
-    """Abstract class for Actions for setting one client property"""
-    def run_on_one_client(self, client, properties):
-        """Set the Client's D-Bus property"""
-        log.debug("D-Bus: %s:%s:%s.Set(%r, %r, %r)", busname,
-                  client.__dbus_object_path__,
-                  dbus.PROPERTIES_IFACE, client_interface,
-                  self.property, self.value_to_set
-                  if not isinstance(self.value_to_set, dbus.Boolean)
-                  else bool(self.value_to_set))
-        client.Set(client_interface, self.property, self.value_to_set,
-                   dbus_interface=dbus.PROPERTIES_IFACE)
-
-class ValueArgumentMixIn(object):
-    """Mixin class for commands taking a value as argument"""
-    def __init__(self, value):
-        self.value_to_set = value
-
-class MillisecondsValueArgumentMixIn(ValueArgumentMixIn):
-    """Mixin class for commands taking a value argument as
-    milliseconds."""
-    @property
-    def value_to_set(self):
-        return self._vts
-    @value_to_set.setter
-    def value_to_set(self, value):
-        """When setting, convert value to a datetime.timedelta"""
-        self._vts = int(round(value.total_seconds() * 1000))
-
-# Actual (non-abstract) command classes
-
-class PrintTableCmd(PrintCmd):
-    def __init__(self, verbose=False):
-        self.verbose = verbose
-
-    def output(self, mandos_objmgr, clients):
-        default_keywords = ("Name", "Enabled", "Timeout", "LastCheckedOK")
-        keywords = default_keywords
-        if self.verbose:
-            keywords = self.all_keywords
-        return str(self.TableOfClients(clients, keywords))
-
-    class TableOfClients(object):
-        tableheaders = {
-            "Name": "Name",
-            "Enabled": "Enabled",
-            "Timeout": "Timeout",
-            "LastCheckedOK": "Last Successful Check",
-            "LastApprovalRequest": "Last Approval Request",
-            "Created": "Created",
-            "Interval": "Interval",
-            "Host": "Host",
-            "Fingerprint": "Fingerprint",
-            "KeyID": "Key ID",
-            "CheckerRunning": "Check Is Running",
-            "LastEnabled": "Last Enabled",
-            "ApprovalPending": "Approval Is Pending",
-            "ApprovedByDefault": "Approved By Default",
-            "ApprovalDelay": "Approval Delay",
-            "ApprovalDuration": "Approval Duration",
-            "Checker": "Checker",
-            "ExtendedTimeout": "Extended Timeout",
-            "Expires": "Expires",
-            "LastCheckerStatus": "Last Checker Status",
-        }
-
-        def __init__(self, clients, keywords, tableheaders=None):
-            self.clients = clients
-            self.keywords = keywords
-            if tableheaders is not None:
-                self.tableheaders = tableheaders
-
-        def __str__(self):
-            return "\n".join(self.rows())
-
-        if sys.version_info.major == 2:
-            __unicode__ = __str__
-            def __str__(self):
-                return str(self).encode(locale.getpreferredencoding())
-
-        def rows(self):
-            format_string = self.row_formatting_string()
-            rows = [self.header_line(format_string)]
-            rows.extend(self.client_line(client, format_string)
-                        for client in self.clients)
-            return rows
-
-        def row_formatting_string(self):
-            "Format string used to format table rows"
-            return " ".join("{{{key}:{width}}}".format(
-                width=max(len(self.tableheaders[key]),
-                          *(len(self.string_from_client(client, key))
-                            for client in self.clients)),
-                key=key)
-                            for key in self.keywords)
-
-        def string_from_client(self, client, key):
-            return self.valuetostring(client[key], key)
-
-        @staticmethod
-        def valuetostring(value, keyword):
-            if isinstance(value, dbus.Boolean):
-                return "Yes" if value else "No"
-            if keyword in ("Timeout", "Interval", "ApprovalDelay",
-                           "ApprovalDuration", "ExtendedTimeout"):
-                return milliseconds_to_string(value)
-            return str(value)
-
-        def header_line(self, format_string):
-            return format_string.format(**self.tableheaders)
-
-        def client_line(self, client, format_string):
-            return format_string.format(
-                **{key: self.string_from_client(client, key)
-                   for key in self.keywords})
-
-
-
-class DumpJSONCmd(PrintCmd):
-    def output(self, mandos_objmgr, clients):
-        data = {client["Name"]:
-                {key: self.dbus_boolean_to_bool(client[key])
-                 for key in self.all_keywords}
-                for client in clients.values()}
-        return json.dumps(data, indent=4, separators=(',', ': '))
-    @staticmethod
-    def dbus_boolean_to_bool(value):
-        if isinstance(value, dbus.Boolean):
-            value = bool(value)
-        return value
-
-class IsEnabledCmd(Command):
-    def run_on_one_client(self, client, properties):
-        if self.is_enabled(client, properties):
-            sys.exit(0)
-        sys.exit(1)
-    def is_enabled(self, client, properties):
-        log.debug("D-Bus: %s:%s:%s.Get(%r, %r)", busname,
-                  client.__dbus_object_path__,
-                  dbus.PROPERTIES_IFACE, client_interface,
-                  "Enabled")
-        return bool(client.Get(client_interface, "Enabled",
-                               dbus_interface=dbus.PROPERTIES_IFACE))
-
-class RemoveCmd(Command):
-    def run_on_one_client(self, client, properties):
-        log.debug("D-Bus: %s:%s:%s.RemoveClient(%r)", busname,
-                  server_path, server_interface,
-                  str(client.__dbus_object_path__))
-        self.mandos.RemoveClient(client.__dbus_object_path__)
-
-class ApproveCmd(Command):
-    def run_on_one_client(self, client, properties):
-        log.debug("D-Bus: %s:%s.Approve(True)",
-                  client.__dbus_object_path__, client_interface)
-        client.Approve(dbus.Boolean(True),
-                       dbus_interface=client_interface)
-
-class DenyCmd(Command):
-    def run_on_one_client(self, client, properties):
-        log.debug("D-Bus: %s:%s.Approve(False)",
-                  client.__dbus_object_path__, client_interface)
-        client.Approve(dbus.Boolean(False),
-                       dbus_interface=client_interface)
-
-class EnableCmd(PropertyCmd):
-    property = "Enabled"
-    value_to_set = dbus.Boolean(True)
-
-class DisableCmd(PropertyCmd):
-    property = "Enabled"
-    value_to_set = dbus.Boolean(False)
-
-class BumpTimeoutCmd(PropertyCmd):
-    property = "LastCheckedOK"
-    value_to_set = ""
-
-class StartCheckerCmd(PropertyCmd):
-    property = "CheckerRunning"
-    value_to_set = dbus.Boolean(True)
-
-class StopCheckerCmd(PropertyCmd):
-    property = "CheckerRunning"
-    value_to_set = dbus.Boolean(False)
-
-class ApproveByDefaultCmd(PropertyCmd):
-    property = "ApprovedByDefault"
-    value_to_set = dbus.Boolean(True)
-
-class DenyByDefaultCmd(PropertyCmd):
-    property = "ApprovedByDefault"
-    value_to_set = dbus.Boolean(False)
-
-class SetCheckerCmd(PropertyCmd, ValueArgumentMixIn):
-    property = "Checker"
-
-class SetHostCmd(PropertyCmd, ValueArgumentMixIn):
-    property = "Host"
-
-class SetSecretCmd(PropertyCmd, ValueArgumentMixIn):
-    @property
-    def value_to_set(self):
-        return self._vts
-    @value_to_set.setter
-    def value_to_set(self, value):
-        """When setting, read data from supplied file object"""
-        self._vts = value.read()
-        value.close()
-    property = "Secret"
-
-class SetTimeoutCmd(PropertyCmd, MillisecondsValueArgumentMixIn):
-    property = "Timeout"
-
-class SetExtendedTimeoutCmd(PropertyCmd,
-                            MillisecondsValueArgumentMixIn):
-    property = "ExtendedTimeout"
-
-class SetIntervalCmd(PropertyCmd, MillisecondsValueArgumentMixIn):
-    property = "Interval"
-
-class SetApprovalDelayCmd(PropertyCmd,
-                          MillisecondsValueArgumentMixIn):
-    property = "ApprovalDelay"
-
-class SetApprovalDurationCmd(PropertyCmd,
-                             MillisecondsValueArgumentMixIn):
-    property = "ApprovalDuration"
-
-def add_command_line_options(parser):
-    parser.add_argument("--version", action="version",
-                        version="%(prog)s {}".format(version),
-                        help="show version number and exit")
-    parser.add_argument("-a", "--all", action="store_true",
-                        help="Select all clients")
-    parser.add_argument("-v", "--verbose", action="store_true",
-                        help="Print all fields")
-    parser.add_argument("-j", "--dump-json", action="store_true",
-                        help="Dump client data in JSON format")
-    enable_disable = parser.add_mutually_exclusive_group()
-    enable_disable.add_argument("-e", "--enable", action="store_true",
-                                help="Enable client")
-    enable_disable.add_argument("-d", "--disable",
-                                action="store_true",
-                                help="disable client")
-    parser.add_argument("-b", "--bump-timeout", action="store_true",
-                        help="Bump timeout for client")
-    start_stop_checker = parser.add_mutually_exclusive_group()
-    start_stop_checker.add_argument("--start-checker",
-                                    action="store_true",
-                                    help="Start checker for client")
-    start_stop_checker.add_argument("--stop-checker",
-                                    action="store_true",
-                                    help="Stop checker for client")
-    parser.add_argument("-V", "--is-enabled", action="store_true",
-                        help="Check if client is enabled")
-    parser.add_argument("-r", "--remove", action="store_true",
-                        help="Remove client")
-    parser.add_argument("-c", "--checker",
-                        help="Set checker command for client")
-    parser.add_argument("-t", "--timeout", type=string_to_delta,
-                        help="Set timeout for client")
-    parser.add_argument("--extended-timeout", type=string_to_delta,
-                        help="Set extended timeout for client")
-    parser.add_argument("-i", "--interval", type=string_to_delta,
-                        help="Set checker interval for client")
-    approve_deny_default = parser.add_mutually_exclusive_group()
-    approve_deny_default.add_argument(
-        "--approve-by-default", action="store_true",
-        default=None, dest="approved_by_default",
-        help="Set client to be approved by default")
-    approve_deny_default.add_argument(
-        "--deny-by-default", action="store_false",
-        dest="approved_by_default",
-        help="Set client to be denied by default")
-    parser.add_argument("--approval-delay", type=string_to_delta,
-                        help="Set delay before client approve/deny")
-    parser.add_argument("--approval-duration", type=string_to_delta,
-                        help="Set duration of one client approval")
-    parser.add_argument("-H", "--host", help="Set host for client")
-    parser.add_argument("-s", "--secret",
-                        type=argparse.FileType(mode="rb"),
-                        help="Set password blob (file) for client")
-    approve_deny = parser.add_mutually_exclusive_group()
-    approve_deny.add_argument(
-        "-A", "--approve", action="store_true",
-        help="Approve any current client request")
-    approve_deny.add_argument("-D", "--deny", action="store_true",
-                              help="Deny any current client request")
-    parser.add_argument("--debug", action="store_true",
-                        help="Debug mode (show D-Bus commands)")
-    parser.add_argument("--check", action="store_true",
-                        help="Run self-test")
-    parser.add_argument("client", nargs="*", help="Client name")
-
-
-def commands_from_options(options):
-
-    commands = []
-
-    if options.dump_json:
-        commands.append(DumpJSONCmd())
-
-    if options.enable:
-        commands.append(EnableCmd())
-
-    if options.disable:
-        commands.append(DisableCmd())
-
-    if options.bump_timeout:
-        commands.append(BumpTimeoutCmd())
-
-    if options.start_checker:
-        commands.append(StartCheckerCmd())
-
-    if options.stop_checker:
-        commands.append(StopCheckerCmd())
-
-    if options.is_enabled:
-        commands.append(IsEnabledCmd())
-
-    if options.checker is not None:
-        commands.append(SetCheckerCmd(options.checker))
-
-    if options.timeout is not None:
-        commands.append(SetTimeoutCmd(options.timeout))
-
-    if options.extended_timeout:
-        commands.append(
-            SetExtendedTimeoutCmd(options.extended_timeout))
-
-    if options.interval is not None:
-        commands.append(SetIntervalCmd(options.interval))
-
-    if options.approved_by_default is not None:
-        if options.approved_by_default:
-            commands.append(ApproveByDefaultCmd())
-        else:
-            commands.append(DenyByDefaultCmd())
-
-    if options.approval_delay is not None:
-        commands.append(SetApprovalDelayCmd(options.approval_delay))
-
-    if options.approval_duration is not None:
-        commands.append(
-            SetApprovalDurationCmd(options.approval_duration))
-
-    if options.host is not None:
-        commands.append(SetHostCmd(options.host))
-
-    if options.secret is not None:
-        commands.append(SetSecretCmd(options.secret))
-
-    if options.approve:
-        commands.append(ApproveCmd())
-
-    if options.deny:
-        commands.append(DenyCmd())
-
-    if options.remove:
-        commands.append(RemoveCmd())
-
-    # If no command option has been given, show table of clients,
-    # optionally verbosely
-    if not commands:
-        commands.append(PrintTableCmd(verbose=options.verbose))
-
-    return commands
-
-
 def check_option_syntax(parser, options):
     """Apply additional restrictions on options, not expressible in
 argparse"""
@@ -707,152 +424,971 @@
         options.remove = True
 
 
-def main():
-    parser = argparse.ArgumentParser()
-
-    add_command_line_options(parser)
-
-    options = parser.parse_args()
-
-    check_option_syntax(parser, options)
-
-    clientnames = options.client
-
-    if options.debug:
-        log.setLevel(logging.DEBUG)
-
+def get_mandos_dbus_object(bus):
+    log.debug("D-Bus: Connect to: (busname=%r, path=%r)",
+              dbus_busname, server_dbus_path)
+    with if_dbus_exception_log_with_exception_and_exit(
+            "Could not connect to Mandos server: %s"):
+        mandos_dbus_object = bus.get_object(dbus_busname,
+                                            server_dbus_path)
+    return mandos_dbus_object
+
+
+@contextlib.contextmanager
+def if_dbus_exception_log_with_exception_and_exit(*args, **kwargs):
     try:
-        bus = dbus.SystemBus()
-        log.debug("D-Bus: Connect to: (name=%r, path=%r)", busname,
-                  server_path)
-        mandos_dbus_objc = bus.get_object(busname, server_path)
-    except dbus.exceptions.DBusException:
-        log.critical("Could not connect to Mandos server")
+        yield
+    except dbus.exceptions.DBusException as e:
+        log.critical(*(args + (e,)), **kwargs)
         sys.exit(1)
 
-    mandos_serv = dbus.Interface(mandos_dbus_objc,
-                                 dbus_interface=server_interface)
-    mandos_serv_object_manager = dbus.Interface(
-        mandos_dbus_objc, dbus_interface=dbus.OBJECT_MANAGER_IFACE)
-
-    # Filter out log message from dbus module
-    dbus_logger = logging.getLogger("dbus.proxies")
+
+def get_managed_objects(object_manager):
+    log.debug("D-Bus: %s:%s:%s.GetManagedObjects()", dbus_busname,
+              server_dbus_path, dbus.OBJECT_MANAGER_IFACE)
+    with if_dbus_exception_log_with_exception_and_exit(
+            "Failed to access Mandos server through D-Bus:\n%s"):
+        with SilenceLogger("dbus.proxies"):
+            managed_objects = object_manager.GetManagedObjects()
+    return managed_objects
+
+
+class SilenceLogger(object):
+    "Simple context manager to silence a particular logger"
+    def __init__(self, loggername):
+        self.logger = logging.getLogger(loggername)
+
+    def __enter__(self):
+        self.logger.addFilter(self.nullfilter)
+        return self
+
     class NullFilter(logging.Filter):
         def filter(self, record):
             return False
-    dbus_filter = NullFilter()
-    try:
-        dbus_logger.addFilter(dbus_filter)
-        log.debug("D-Bus: %s:%s:%s.GetManagedObjects()", busname,
-                  server_path, dbus.OBJECT_MANAGER_IFACE)
-        mandos_clients = {path: ifs_and_props[client_interface]
-                          for path, ifs_and_props in
-                          mandos_serv_object_manager
-                          .GetManagedObjects().items()
-                          if client_interface in ifs_and_props}
-    except dbus.exceptions.DBusException as e:
-        log.critical("Failed to access Mandos server through D-Bus:"
-                     "\n%s", e)
-        sys.exit(1)
-    finally:
-        # restore dbus logger
-        dbus_logger.removeFilter(dbus_filter)
-
-    # Compile dict of (clients: properties) to process
-    clients = {}
-
-    if not clientnames:
-        clients = {(log.debug("D-Bus: Connect to: (name=%r, path=%r)",
-                              busname, str(path)) and False) or
-                   bus.get_object(busname, path): properties
-                   for path, properties in mandos_clients.items()}
-    else:
-        for name in clientnames:
-            for path, client in mandos_clients.items():
-                if client["Name"] == name:
-                    log.debug("D-Bus: Connect to: (name=%r, path=%r)",
-                              busname, str(path))
-                    client_objc = bus.get_object(busname, path)
-                    clients[client_objc] = client
-                    break
-            else:
-                log.critical("Client not found on server: %r", name)
-                sys.exit(1)
-
-    # Run all commands on clients
-    commands = commands_from_options(options)
-    for command in commands:
-        command.run(mandos_serv, mandos_serv_object_manager, clients)
+
+    nullfilter = NullFilter()
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.logger.removeFilter(self.nullfilter)
+
+
+def commands_from_options(options):
+
+    commands = []
+
+    if options.is_enabled:
+        commands.append(command.IsEnabled())
+
+    if options.approve:
+        commands.append(command.Approve())
+
+    if options.deny:
+        commands.append(command.Deny())
+
+    if options.remove:
+        commands.append(command.Remove())
+
+    if options.dump_json:
+        commands.append(command.DumpJSON())
+
+    if options.enable:
+        commands.append(command.Enable())
+
+    if options.disable:
+        commands.append(command.Disable())
+
+    if options.bump_timeout:
+        commands.append(command.BumpTimeout())
+
+    if options.start_checker:
+        commands.append(command.StartChecker())
+
+    if options.stop_checker:
+        commands.append(command.StopChecker())
+
+    if options.approved_by_default is not None:
+        if options.approved_by_default:
+            commands.append(command.ApproveByDefault())
+        else:
+            commands.append(command.DenyByDefault())
+
+    if options.checker is not None:
+        commands.append(command.SetChecker(options.checker))
+
+    if options.host is not None:
+        commands.append(command.SetHost(options.host))
+
+    if options.secret is not None:
+        commands.append(command.SetSecret(options.secret))
+
+    if options.timeout is not None:
+        commands.append(command.SetTimeout(options.timeout))
+
+    if options.extended_timeout:
+        commands.append(
+            command.SetExtendedTimeout(options.extended_timeout))
+
+    if options.interval is not None:
+        commands.append(command.SetInterval(options.interval))
+
+    if options.approval_delay is not None:
+        commands.append(
+            command.SetApprovalDelay(options.approval_delay))
+
+    if options.approval_duration is not None:
+        commands.append(
+            command.SetApprovalDuration(options.approval_duration))
+
+    # If no command option has been given, show table of clients,
+    # optionally verbosely
+    if not commands:
+        commands.append(command.PrintTable(verbose=options.verbose))
+
+    return commands
+
+
+class command(object):
+    """A namespace for command classes"""
+
+    class Base(object):
+        """Abstract base class for commands"""
+        def run(self, clients, bus=None, mandos=None):
+            """Normal commands should implement run_on_one_client(),
+but commands which want to operate on all clients at the same time can
+override this run() method instead.
+"""
+            self.mandos = mandos
+            for clientpath, properties in clients.items():
+                log.debug("D-Bus: Connect to: (busname=%r, path=%r)",
+                          dbus_busname, str(clientpath))
+                client = bus.get_object(dbus_busname, clientpath)
+                self.run_on_one_client(client, properties)
+
+
+    class IsEnabled(Base):
+        def run(self, clients, bus=None, mandos=None):
+            client, properties = next(iter(clients.items()))
+            if self.is_enabled(client, properties):
+                sys.exit(0)
+            sys.exit(1)
+        def is_enabled(self, client, properties):
+            return properties["Enabled"]
+
+
+    class Approve(Base):
+        def run_on_one_client(self, client, properties):
+            log.debug("D-Bus: %s:%s:%s.Approve(True)", dbus_busname,
+                      client.__dbus_object_path__,
+                      client_dbus_interface)
+            client.Approve(dbus.Boolean(True),
+                           dbus_interface=client_dbus_interface)
+
+
+    class Deny(Base):
+        def run_on_one_client(self, client, properties):
+            log.debug("D-Bus: %s:%s:%s.Approve(False)", dbus_busname,
+                      client.__dbus_object_path__,
+                      client_dbus_interface)
+            client.Approve(dbus.Boolean(False),
+                           dbus_interface=client_dbus_interface)
+
+
+    class Remove(Base):
+        def run_on_one_client(self, client, properties):
+            log.debug("D-Bus: %s:%s:%s.RemoveClient(%r)",
+                      dbus_busname, server_dbus_path,
+                      server_dbus_interface,
+                      str(client.__dbus_object_path__))
+            self.mandos.RemoveClient(client.__dbus_object_path__)
+
+
+    class Output(Base):
+        """Abstract class for commands outputting client details"""
+        all_keywords = ("Name", "Enabled", "Timeout", "LastCheckedOK",
+                        "Created", "Interval", "Host", "KeyID",
+                        "Fingerprint", "CheckerRunning",
+                        "LastEnabled", "ApprovalPending",
+                        "ApprovedByDefault", "LastApprovalRequest",
+                        "ApprovalDelay", "ApprovalDuration",
+                        "Checker", "ExtendedTimeout", "Expires",
+                        "LastCheckerStatus")
+
+        def run(self, clients, bus=None, mandos=None):
+            print(self.output(clients.values()))
+
+        def output(self, clients):
+            raise NotImplementedError()
+
+
+    class DumpJSON(Output):
+        def output(self, clients):
+            data = {client["Name"]:
+                    {key: self.dbus_boolean_to_bool(client[key])
+                     for key in self.all_keywords}
+                    for client in clients}
+            return json.dumps(data, indent=4, separators=(',', ': '))
+
+        @staticmethod
+        def dbus_boolean_to_bool(value):
+            if isinstance(value, dbus.Boolean):
+                value = bool(value)
+            return value
+
+
+    class PrintTable(Output):
+        def __init__(self, verbose=False):
+            self.verbose = verbose
+
+        def output(self, clients):
+            default_keywords = ("Name", "Enabled", "Timeout",
+                                "LastCheckedOK")
+            keywords = default_keywords
+            if self.verbose:
+                keywords = self.all_keywords
+            return str(self.TableOfClients(clients, keywords))
+
+        class TableOfClients(object):
+            tableheaders = {
+                "Name": "Name",
+                "Enabled": "Enabled",
+                "Timeout": "Timeout",
+                "LastCheckedOK": "Last Successful Check",
+                "LastApprovalRequest": "Last Approval Request",
+                "Created": "Created",
+                "Interval": "Interval",
+                "Host": "Host",
+                "Fingerprint": "Fingerprint",
+                "KeyID": "Key ID",
+                "CheckerRunning": "Check Is Running",
+                "LastEnabled": "Last Enabled",
+                "ApprovalPending": "Approval Is Pending",
+                "ApprovedByDefault": "Approved By Default",
+                "ApprovalDelay": "Approval Delay",
+                "ApprovalDuration": "Approval Duration",
+                "Checker": "Checker",
+                "ExtendedTimeout": "Extended Timeout",
+                "Expires": "Expires",
+                "LastCheckerStatus": "Last Checker Status",
+            }
+
+            def __init__(self, clients, keywords):
+                self.clients = clients
+                self.keywords = keywords
+
+            def __str__(self):
+                return "\n".join(self.rows())
+
+            if sys.version_info.major == 2:
+                __unicode__ = __str__
+                def __str__(self):
+                    return str(self).encode(
+                        locale.getpreferredencoding())
+
+            def rows(self):
+                format_string = self.row_formatting_string()
+                rows = [self.header_line(format_string)]
+                rows.extend(self.client_line(client, format_string)
+                            for client in self.clients)
+                return rows
+
+            def row_formatting_string(self):
+                "Format string used to format table rows"
+                return " ".join("{{{key}:{width}}}".format(
+                    width=max(len(self.tableheaders[key]),
+                              *(len(self.string_from_client(client,
+                                                            key))
+                                for client in self.clients)),
+                    key=key)
+                                for key in self.keywords)
+
+            def string_from_client(self, client, key):
+                return self.valuetostring(client[key], key)
+
+            @classmethod
+            def valuetostring(cls, value, keyword):
+                if isinstance(value, dbus.Boolean):
+                    return "Yes" if value else "No"
+                if keyword in ("Timeout", "Interval", "ApprovalDelay",
+                               "ApprovalDuration", "ExtendedTimeout"):
+                    return cls.milliseconds_to_string(value)
+                return str(value)
+
+            def header_line(self, format_string):
+                return format_string.format(**self.tableheaders)
+
+            def client_line(self, client, format_string):
+                return format_string.format(
+                    **{key: self.string_from_client(client, key)
+                       for key in self.keywords})
+
+            @staticmethod
+            def milliseconds_to_string(ms):
+                td = datetime.timedelta(0, 0, 0, ms)
+                return ("{days}{hours:02}:{minutes:02}:{seconds:02}"
+                        .format(days="{}T".format(td.days)
+                                if td.days else "",
+                                hours=td.seconds // 3600,
+                                minutes=(td.seconds % 3600) // 60,
+                                seconds=td.seconds % 60))
+
+
+    class Property(Base):
+        "Abstract class for Actions for setting one client property"
+
+        def run_on_one_client(self, client, properties):
+            """Set the Client's D-Bus property"""
+            log.debug("D-Bus: %s:%s:%s.Set(%r, %r, %r)", dbus_busname,
+                      client.__dbus_object_path__,
+                      dbus.PROPERTIES_IFACE, client_dbus_interface,
+                      self.propname, self.value_to_set
+                      if not isinstance(self.value_to_set,
+                                        dbus.Boolean)
+                      else bool(self.value_to_set))
+            client.Set(client_dbus_interface, self.propname,
+                       self.value_to_set,
+                       dbus_interface=dbus.PROPERTIES_IFACE)
+
+        @property
+        def propname(self):
+            raise NotImplementedError()
+
+
+    class Enable(Property):
+        propname = "Enabled"
+        value_to_set = dbus.Boolean(True)
+
+
+    class Disable(Property):
+        propname = "Enabled"
+        value_to_set = dbus.Boolean(False)
+
+
+    class BumpTimeout(Property):
+        propname = "LastCheckedOK"
+        value_to_set = ""
+
+
+    class StartChecker(Property):
+        propname = "CheckerRunning"
+        value_to_set = dbus.Boolean(True)
+
+
+    class StopChecker(Property):
+        propname = "CheckerRunning"
+        value_to_set = dbus.Boolean(False)
+
+
+    class ApproveByDefault(Property):
+        propname = "ApprovedByDefault"
+        value_to_set = dbus.Boolean(True)
+
+
+    class DenyByDefault(Property):
+        propname = "ApprovedByDefault"
+        value_to_set = dbus.Boolean(False)
+
+
+    class PropertyValue(Property):
+        "Abstract class for Property recieving a value as argument"
+        def __init__(self, value):
+            self.value_to_set = value
+
+
+    class SetChecker(PropertyValue):
+        propname = "Checker"
+
+
+    class SetHost(PropertyValue):
+        propname = "Host"
+
+
+    class SetSecret(PropertyValue):
+        propname = "Secret"
+
+        @property
+        def value_to_set(self):
+            return self._vts
+
+        @value_to_set.setter
+        def value_to_set(self, value):
+            """When setting, read data from supplied file object"""
+            self._vts = value.read()
+            value.close()
+
+
+    class MillisecondsPropertyValueArgument(PropertyValue):
+        """Abstract class for PropertyValue taking a value argument as
+a datetime.timedelta() but should store it as milliseconds."""
+
+        @property
+        def value_to_set(self):
+            return self._vts
+
+        @value_to_set.setter
+        def value_to_set(self, value):
+            "When setting, convert value from a datetime.timedelta"
+            self._vts = int(round(value.total_seconds() * 1000))
+
+
+    class SetTimeout(MillisecondsPropertyValueArgument):
+        propname = "Timeout"
+
+
+    class SetExtendedTimeout(MillisecondsPropertyValueArgument):
+        propname = "ExtendedTimeout"
+
+
+    class SetInterval(MillisecondsPropertyValueArgument):
+        propname = "Interval"
+
+
+    class SetApprovalDelay(MillisecondsPropertyValueArgument):
+        propname = "ApprovalDelay"
+
+
+    class SetApprovalDuration(MillisecondsPropertyValueArgument):
+        propname = "ApprovalDuration"
+
 
 
-class Test_milliseconds_to_string(unittest.TestCase):
-    def test_all(self):
-        self.assertEqual(milliseconds_to_string(93785000),
-                         "1T02:03:05")
-    def test_no_days(self):
-        self.assertEqual(milliseconds_to_string(7385000), "02:03:05")
-    def test_all_zero(self):
-        self.assertEqual(milliseconds_to_string(0), "00:00:00")
-    def test_no_fractional_seconds(self):
-        self.assertEqual(milliseconds_to_string(400), "00:00:00")
-        self.assertEqual(milliseconds_to_string(900), "00:00:00")
-        self.assertEqual(milliseconds_to_string(1900), "00:00:01")
-
-class Test_string_to_delta(unittest.TestCase):
-    def test_handles_basic_rfc3339(self):
+class TestCaseWithAssertLogs(unittest.TestCase):
+    """unittest.TestCase.assertLogs only exists in Python 3.4"""
+
+    if not hasattr(unittest.TestCase, "assertLogs"):
+        @contextlib.contextmanager
+        def assertLogs(self, logger, level=logging.INFO):
+            capturing_handler = self.CapturingLevelHandler(level)
+            old_level = logger.level
+            old_propagate = logger.propagate
+            logger.addHandler(capturing_handler)
+            logger.setLevel(level)
+            logger.propagate = False
+            try:
+                yield capturing_handler.watcher
+            finally:
+                logger.propagate = old_propagate
+                logger.removeHandler(capturing_handler)
+                logger.setLevel(old_level)
+            self.assertGreater(len(capturing_handler.watcher.records),
+                               0)
+
+        class CapturingLevelHandler(logging.Handler):
+            def __init__(self, level, *args, **kwargs):
+                logging.Handler.__init__(self, *args, **kwargs)
+                self.watcher = self.LoggingWatcher([], [])
+            def emit(self, record):
+                self.watcher.records.append(record)
+                self.watcher.output.append(self.format(record))
+
+            LoggingWatcher = collections.namedtuple("LoggingWatcher",
+                                                    ("records",
+                                                     "output"))
+
+
+class Test_string_to_delta(TestCaseWithAssertLogs):
+    # Just test basic RFC 3339 functionality here, the doc string for
+    # rfc3339_duration_to_delta() already has more comprehensive
+    # tests, which is run by doctest.
+
+    def test_rfc3339_zero_seconds(self):
         self.assertEqual(string_to_delta("PT0S"),
                          datetime.timedelta())
+
+    def test_rfc3339_zero_days(self):
         self.assertEqual(string_to_delta("P0D"),
                          datetime.timedelta())
+
+    def test_rfc3339_one_second(self):
         self.assertEqual(string_to_delta("PT1S"),
                          datetime.timedelta(0, 1))
+
+    def test_rfc3339_two_hours(self):
         self.assertEqual(string_to_delta("PT2H"),
                          datetime.timedelta(0, 7200))
+
     def test_falls_back_to_pre_1_6_1_with_warning(self):
-        # assertLogs only exists in Python 3.4
-        if hasattr(self, "assertLogs"):
-            with self.assertLogs(log, logging.WARNING):
-                value = string_to_delta("2h")
-        else:
-            class WarningFilter(logging.Filter):
-                """Don't show, but record the presence of, warnings"""
-                def filter(self, record):
-                    is_warning = record.levelno >= logging.WARNING
-                    self.found = is_warning or getattr(self, "found",
-                                                       False)
-                    return not is_warning
-            warning_filter = WarningFilter()
-            log.addFilter(warning_filter)
-            try:
-                value = string_to_delta("2h")
-            finally:
-                log.removeFilter(warning_filter)
-            self.assertTrue(getattr(warning_filter, "found", False))
+        with self.assertLogs(log, logging.WARNING):
+            value = string_to_delta("2h")
         self.assertEqual(value, datetime.timedelta(0, 7200))
 
 
-class TestCmd(unittest.TestCase):
+class Test_check_option_syntax(unittest.TestCase):
+    def setUp(self):
+        self.parser = argparse.ArgumentParser()
+        add_command_line_options(self.parser)
+
+    def test_actions_requires_client_or_all(self):
+        for action, value in self.actions.items():
+            options = self.parser.parse_args()
+            setattr(options, action, value)
+            with self.assertParseError():
+                self.check_option_syntax(options)
+
+    # This mostly corresponds to the definition from has_actions() in
+    # check_option_syntax()
+    actions = {
+        # The actual values set here are not that important, but we do
+        # at least stick to the correct types, even though they are
+        # never used
+        "enable": True,
+        "disable": True,
+        "bump_timeout": True,
+        "start_checker": True,
+        "stop_checker": True,
+        "is_enabled": True,
+        "remove": True,
+        "checker": "x",
+        "timeout": datetime.timedelta(),
+        "extended_timeout": datetime.timedelta(),
+        "interval": datetime.timedelta(),
+        "approved_by_default": True,
+        "approval_delay": datetime.timedelta(),
+        "approval_duration": datetime.timedelta(),
+        "host": "x",
+        "secret": io.BytesIO(b"x"),
+        "approve": True,
+        "deny": True,
+    }
+
+    @contextlib.contextmanager
+    def assertParseError(self):
+        with self.assertRaises(SystemExit) as e:
+            with self.redirect_stderr_to_devnull():
+                yield
+        # Exit code from argparse is guaranteed to be "2".  Reference:
+        # https://docs.python.org/3/library
+        # /argparse.html#exiting-methods
+        self.assertEqual(e.exception.code, 2)
+
+    @staticmethod
+    @contextlib.contextmanager
+    def redirect_stderr_to_devnull():
+        null = os.open(os.path.devnull, os.O_RDWR)
+        stderrcopy = os.dup(sys.stderr.fileno())
+        os.dup2(null, sys.stderr.fileno())
+        os.close(null)
+        try:
+            yield
+        finally:
+            # restore stderr
+            os.dup2(stderrcopy, sys.stderr.fileno())
+            os.close(stderrcopy)
+
+    def check_option_syntax(self, options):
+        check_option_syntax(self.parser, options)
+
+    def test_actions_all_conflicts_with_verbose(self):
+        for action, value in self.actions.items():
+            options = self.parser.parse_args()
+            setattr(options, action, value)
+            options.all = True
+            options.verbose = True
+            with self.assertParseError():
+                self.check_option_syntax(options)
+
+    def test_actions_with_client_conflicts_with_verbose(self):
+        for action, value in self.actions.items():
+            options = self.parser.parse_args()
+            setattr(options, action, value)
+            options.verbose = True
+            options.client = ["foo"]
+            with self.assertParseError():
+                self.check_option_syntax(options)
+
+    def test_dump_json_conflicts_with_verbose(self):
+        options = self.parser.parse_args()
+        options.dump_json = True
+        options.verbose = True
+        with self.assertParseError():
+            self.check_option_syntax(options)
+
+    def test_dump_json_conflicts_with_action(self):
+        for action, value in self.actions.items():
+            options = self.parser.parse_args()
+            setattr(options, action, value)
+            options.dump_json = True
+            with self.assertParseError():
+                self.check_option_syntax(options)
+
+    def test_all_can_not_be_alone(self):
+        options = self.parser.parse_args()
+        options.all = True
+        with self.assertParseError():
+            self.check_option_syntax(options)
+
+    def test_all_is_ok_with_any_action(self):
+        for action, value in self.actions.items():
+            options = self.parser.parse_args()
+            setattr(options, action, value)
+            options.all = True
+            self.check_option_syntax(options)
+
+    def test_any_action_is_ok_with_one_client(self):
+        for action, value in self.actions.items():
+            options = self.parser.parse_args()
+            setattr(options, action, value)
+            options.client = ["foo"]
+            self.check_option_syntax(options)
+
+    def test_one_client_with_all_actions_except_is_enabled(self):
+        options = self.parser.parse_args()
+        for action, value in self.actions.items():
+            if action == "is_enabled":
+                continue
+            setattr(options, action, value)
+        options.client = ["foo"]
+        self.check_option_syntax(options)
+
+    def test_two_clients_with_all_actions_except_is_enabled(self):
+        options = self.parser.parse_args()
+        for action, value in self.actions.items():
+            if action == "is_enabled":
+                continue
+            setattr(options, action, value)
+        options.client = ["foo", "barbar"]
+        self.check_option_syntax(options)
+
+    def test_two_clients_are_ok_with_actions_except_is_enabled(self):
+        for action, value in self.actions.items():
+            if action == "is_enabled":
+                continue
+            options = self.parser.parse_args()
+            setattr(options, action, value)
+            options.client = ["foo", "barbar"]
+            self.check_option_syntax(options)
+
+    def test_is_enabled_fails_without_client(self):
+        options = self.parser.parse_args()
+        options.is_enabled = True
+        with self.assertParseError():
+            self.check_option_syntax(options)
+
+    def test_is_enabled_fails_with_two_clients(self):
+        options = self.parser.parse_args()
+        options.is_enabled = True
+        options.client = ["foo", "barbar"]
+        with self.assertParseError():
+            self.check_option_syntax(options)
+
+    def test_remove_can_only_be_combined_with_action_deny(self):
+        for action, value in self.actions.items():
+            if action in {"remove", "deny"}:
+                continue
+            options = self.parser.parse_args()
+            setattr(options, action, value)
+            options.all = True
+            options.remove = True
+            with self.assertParseError():
+                self.check_option_syntax(options)
+
+
+class Test_get_mandos_dbus_object(TestCaseWithAssertLogs):
+    def test_calls_and_returns_get_object_on_bus(self):
+        class MockBus(object):
+            called = False
+            def get_object(mockbus_self, busname, dbus_path):
+                # Note that "self" is still the testcase instance,
+                # this MockBus instance is in "mockbus_self".
+                self.assertEqual(busname, dbus_busname)
+                self.assertEqual(dbus_path, server_dbus_path)
+                mockbus_self.called = True
+                return mockbus_self
+
+        mockbus = get_mandos_dbus_object(bus=MockBus())
+        self.assertIsInstance(mockbus, MockBus)
+        self.assertTrue(mockbus.called)
+
+    def test_logs_and_exits_on_dbus_error(self):
+        class MockBusFailing(object):
+            def get_object(self, busname, dbus_path):
+                raise dbus.exceptions.DBusException("Test")
+
+        with self.assertLogs(log, logging.CRITICAL):
+            with self.assertRaises(SystemExit) as e:
+                bus = get_mandos_dbus_object(bus=MockBusFailing())
+
+        if isinstance(e.exception.code, int):
+            self.assertNotEqual(e.exception.code, 0)
+        else:
+            self.assertIsNotNone(e.exception.code)
+
+
+class Test_get_managed_objects(TestCaseWithAssertLogs):
+    def test_calls_and_returns_GetManagedObjects(self):
+        managed_objects = {"/clients/foo": { "Name": "foo"}}
+        class MockObjectManager(object):
+            def GetManagedObjects(self):
+                return managed_objects
+        retval = get_managed_objects(MockObjectManager())
+        self.assertDictEqual(managed_objects, retval)
+
+    def test_logs_and_exits_on_dbus_error(self):
+        dbus_logger = logging.getLogger("dbus.proxies")
+
+        class MockObjectManagerFailing(object):
+            def GetManagedObjects(self):
+                dbus_logger.error("Test")
+                raise dbus.exceptions.DBusException("Test")
+
+        class CountingHandler(logging.Handler):
+            count = 0
+            def emit(self, record):
+                self.count += 1
+
+        counting_handler = CountingHandler()
+
+        dbus_logger.addHandler(counting_handler)
+
+        try:
+            with self.assertLogs(log, logging.CRITICAL) as watcher:
+                with self.assertRaises(SystemExit) as e:
+                    get_managed_objects(MockObjectManagerFailing())
+        finally:
+            dbus_logger.removeFilter(counting_handler)
+
+        # Make sure the dbus logger was suppressed
+        self.assertEqual(counting_handler.count, 0)
+
+        # Test that the dbus_logger still works
+        with self.assertLogs(dbus_logger, logging.ERROR):
+            dbus_logger.error("Test")
+
+        if isinstance(e.exception.code, int):
+            self.assertNotEqual(e.exception.code, 0)
+        else:
+            self.assertIsNotNone(e.exception.code)
+
+
+class Test_commands_from_options(unittest.TestCase):
+    def setUp(self):
+        self.parser = argparse.ArgumentParser()
+        add_command_line_options(self.parser)
+
+    def test_is_enabled(self):
+        self.assert_command_from_args(["--is-enabled", "foo"],
+                                      command.IsEnabled)
+
+    def assert_command_from_args(self, args, command_cls,
+                                 **cmd_attrs):
+        """Assert that parsing ARGS should result in an instance of
+COMMAND_CLS with (optionally) all supplied attributes (CMD_ATTRS)."""
+        options = self.parser.parse_args(args)
+        check_option_syntax(self.parser, options)
+        commands = commands_from_options(options)
+        self.assertEqual(len(commands), 1)
+        command = commands[0]
+        self.assertIsInstance(command, command_cls)
+        for key, value in cmd_attrs.items():
+            self.assertEqual(getattr(command, key), value)
+
+    def test_is_enabled_short(self):
+        self.assert_command_from_args(["-V", "foo"],
+                                      command.IsEnabled)
+
+    def test_approve(self):
+        self.assert_command_from_args(["--approve", "foo"],
+                                      command.Approve)
+
+    def test_approve_short(self):
+        self.assert_command_from_args(["-A", "foo"], command.Approve)
+
+    def test_deny(self):
+        self.assert_command_from_args(["--deny", "foo"], command.Deny)
+
+    def test_deny_short(self):
+        self.assert_command_from_args(["-D", "foo"], command.Deny)
+
+    def test_remove(self):
+        self.assert_command_from_args(["--remove", "foo"],
+                                      command.Remove)
+
+    def test_deny_before_remove(self):
+        options = self.parser.parse_args(["--deny", "--remove",
+                                          "foo"])
+        check_option_syntax(self.parser, options)
+        commands = commands_from_options(options)
+        self.assertEqual(len(commands), 2)
+        self.assertIsInstance(commands[0], command.Deny)
+        self.assertIsInstance(commands[1], command.Remove)
+
+    def test_deny_before_remove_reversed(self):
+        options = self.parser.parse_args(["--remove", "--deny",
+                                          "--all"])
+        check_option_syntax(self.parser, options)
+        commands = commands_from_options(options)
+        self.assertEqual(len(commands), 2)
+        self.assertIsInstance(commands[0], command.Deny)
+        self.assertIsInstance(commands[1], command.Remove)
+
+    def test_remove_short(self):
+        self.assert_command_from_args(["-r", "foo"], command.Remove)
+
+    def test_dump_json(self):
+        self.assert_command_from_args(["--dump-json"],
+                                      command.DumpJSON)
+
+    def test_enable(self):
+        self.assert_command_from_args(["--enable", "foo"],
+                                      command.Enable)
+
+    def test_enable_short(self):
+        self.assert_command_from_args(["-e", "foo"], command.Enable)
+
+    def test_disable(self):
+        self.assert_command_from_args(["--disable", "foo"],
+                                      command.Disable)
+
+    def test_disable_short(self):
+        self.assert_command_from_args(["-d", "foo"], command.Disable)
+
+    def test_bump_timeout(self):
+        self.assert_command_from_args(["--bump-timeout", "foo"],
+                                      command.BumpTimeout)
+
+    def test_bump_timeout_short(self):
+        self.assert_command_from_args(["-b", "foo"],
+                                      command.BumpTimeout)
+
+    def test_start_checker(self):
+        self.assert_command_from_args(["--start-checker", "foo"],
+                                      command.StartChecker)
+
+    def test_stop_checker(self):
+        self.assert_command_from_args(["--stop-checker", "foo"],
+                                      command.StopChecker)
+
+    def test_approve_by_default(self):
+        self.assert_command_from_args(["--approve-by-default", "foo"],
+                                      command.ApproveByDefault)
+
+    def test_deny_by_default(self):
+        self.assert_command_from_args(["--deny-by-default", "foo"],
+                                      command.DenyByDefault)
+
+    def test_checker(self):
+        self.assert_command_from_args(["--checker", ":", "foo"],
+                                      command.SetChecker,
+                                      value_to_set=":")
+
+    def test_checker_empty(self):
+        self.assert_command_from_args(["--checker", "", "foo"],
+                                      command.SetChecker,
+                                      value_to_set="")
+
+    def test_checker_short(self):
+        self.assert_command_from_args(["-c", ":", "foo"],
+                                      command.SetChecker,
+                                      value_to_set=":")
+
+    def test_host(self):
+        self.assert_command_from_args(["--host", "foo.example.org",
+                                       "foo"], command.SetHost,
+                                      value_to_set="foo.example.org")
+
+    def test_host_short(self):
+        self.assert_command_from_args(["-H", "foo.example.org",
+                                       "foo"], command.SetHost,
+                                      value_to_set="foo.example.org")
+
+    def test_secret_devnull(self):
+        self.assert_command_from_args(["--secret", os.path.devnull,
+                                       "foo"], command.SetSecret,
+                                      value_to_set=b"")
+
+    def test_secret_tempfile(self):
+        with tempfile.NamedTemporaryFile(mode="r+b") as f:
+            value = b"secret\0xyzzy\nbar"
+            f.write(value)
+            f.seek(0)
+            self.assert_command_from_args(["--secret", f.name,
+                                           "foo"], command.SetSecret,
+                                          value_to_set=value)
+
+    def test_secret_devnull_short(self):
+        self.assert_command_from_args(["-s", os.path.devnull, "foo"],
+                                      command.SetSecret,
+                                      value_to_set=b"")
+
+    def test_secret_tempfile_short(self):
+        with tempfile.NamedTemporaryFile(mode="r+b") as f:
+            value = b"secret\0xyzzy\nbar"
+            f.write(value)
+            f.seek(0)
+            self.assert_command_from_args(["-s", f.name, "foo"],
+                                          command.SetSecret,
+                                          value_to_set=value)
+
+    def test_timeout(self):
+        self.assert_command_from_args(["--timeout", "PT5M", "foo"],
+                                      command.SetTimeout,
+                                      value_to_set=300000)
+
+    def test_timeout_short(self):
+        self.assert_command_from_args(["-t", "PT5M", "foo"],
+                                      command.SetTimeout,
+                                      value_to_set=300000)
+
+    def test_extended_timeout(self):
+        self.assert_command_from_args(["--extended-timeout", "PT15M",
+                                       "foo"],
+                                      command.SetExtendedTimeout,
+                                      value_to_set=900000)
+
+    def test_interval(self):
+        self.assert_command_from_args(["--interval", "PT2M", "foo"],
+                                      command.SetInterval,
+                                      value_to_set=120000)
+
+    def test_interval_short(self):
+        self.assert_command_from_args(["-i", "PT2M", "foo"],
+                                      command.SetInterval,
+                                      value_to_set=120000)
+
+    def test_approval_delay(self):
+        self.assert_command_from_args(["--approval-delay", "PT30S",
+                                       "foo"],
+                                      command.SetApprovalDelay,
+                                      value_to_set=30000)
+
+    def test_approval_duration(self):
+        self.assert_command_from_args(["--approval-duration", "PT1S",
+                                       "foo"],
+                                      command.SetApprovalDuration,
+                                      value_to_set=1000)
+
+    def test_print_table(self):
+        self.assert_command_from_args([], command.PrintTable,
+                                      verbose=False)
+
+    def test_print_table_verbose(self):
+        self.assert_command_from_args(["--verbose"],
+                                      command.PrintTable,
+                                      verbose=True)
+
+    def test_print_table_verbose_short(self):
+        self.assert_command_from_args(["-v"], command.PrintTable,
+                                      verbose=True)
+
+
+class TestCommand(unittest.TestCase):
     """Abstract class for tests of command classes"""
+
     def setUp(self):
         testcase = self
         class MockClient(object):
             def __init__(self, name, **attributes):
-                self.__dbus_object_path__ = "objpath_{}".format(name)
+                self.__dbus_object_path__ = "/clients/{}".format(name)
                 self.attributes = attributes
                 self.attributes["Name"] = name
                 self.calls = []
-            def Set(self, interface, property, value, dbus_interface):
-                testcase.assertEqual(interface, client_interface)
-                testcase.assertEqual(dbus_interface,
-                                     dbus.PROPERTIES_IFACE)
-                self.attributes[property] = value
-            def Get(self, interface, property, dbus_interface):
-                testcase.assertEqual(interface, client_interface)
-                testcase.assertEqual(dbus_interface,
-                                     dbus.PROPERTIES_IFACE)
-                return self.attributes[property]
+            def Set(self, interface, propname, value, dbus_interface):
+                testcase.assertEqual(interface, client_dbus_interface)
+                testcase.assertEqual(dbus_interface,
+                                     dbus.PROPERTIES_IFACE)
+                self.attributes[propname] = value
             def Approve(self, approve, dbus_interface):
-                testcase.assertEqual(dbus_interface, client_interface)
+                testcase.assertEqual(dbus_interface,
+                                     client_dbus_interface)
                 self.calls.append(("Approve", (approve,
                                                dbus_interface)))
         self.client = MockClient(
@@ -898,180 +1434,261 @@
             ApprovedByDefault=dbus.Boolean(False),
             LastApprovalRequest="2019-01-03T00:00:00",
             ApprovalDelay=30000,
-            ApprovalDuration=1000,
+            ApprovalDuration=93785000,
             Checker=":",
             ExtendedTimeout=900000,
             Expires="2019-02-05T00:00:00",
             LastCheckerStatus=-2)
         self.clients =  collections.OrderedDict(
             [
-                (self.client, self.client.attributes),
-                (self.other_client, self.other_client.attributes),
+                ("/clients/foo", self.client.attributes),
+                ("/clients/barbar", self.other_client.attributes),
             ])
-        self.one_client = {self.client: self.client.attributes}
-
-class TestPrintTableCmd(TestCmd):
-    def test_normal(self):
-        output = PrintTableCmd().output(None, self.clients.values())
-        expected_output = """
-Name   Enabled Timeout  Last Successful Check
-foo    Yes     00:05:00 2019-02-03T00:00:00  
-barbar Yes     00:05:00 2019-02-04T00:00:00  
-"""[1:-1]
-        self.assertEqual(output, expected_output)
-    def test_verbose(self):
-        output = PrintTableCmd(verbose=True).output(
-            None, self.clients.values())
-        expected_output = """
-Name   Enabled Timeout  Last Successful Check Created             Interval Host            Key ID                                                           Fingerprint                              Check Is Running Last Enabled        Approval Is Pending Approved By Default Last Approval Request Approval Delay Approval Duration Checker              Extended Timeout Expires             Last Checker Status
-foo    Yes     00:05:00 2019-02-03T00:00:00   2019-01-02T00:00:00 00:02:00 foo.example.org 92ed150794387c03ce684574b1139a6594a34f895daaaf09fd8ea90a27cddb12 778827225BA7DE539C5A7CFA59CFF7CDBD9A5920 No               2019-01-03T00:00:00 No                  Yes                                       00:00:00       00:00:01          fping -q -- %(host)s 00:15:00         2019-02-04T00:00:00 0                  
-barbar Yes     00:05:00 2019-02-04T00:00:00   2019-01-03T00:00:00 00:02:00 192.0.2.3       0558568eedd67d622f5c83b35a115f796ab612cff5ad227247e46c2b020f441c 3E393AEAEFB84C7E89E2F547B3A107558FCA3A27 Yes              2019-01-04T00:00:00 No                  No                  2019-01-03T00:00:00   00:00:30       00:00:01          :                    00:15:00         2019-02-05T00:00:00 -2                 
-"""[1:-1]
-        self.assertEqual(output, expected_output)
-    def test_one_client(self):
-        output = PrintTableCmd().output(None, self.one_client.values())
-        expected_output = """
-Name Enabled Timeout  Last Successful Check
-foo  Yes     00:05:00 2019-02-03T00:00:00  
-"""[1:-1]
-        self.assertEqual(output, expected_output)
-
-class TestDumpJSONCmd(TestCmd):
-    def setUp(self):
-        self.expected_json = {
-            "foo": {
-                "Name": "foo",
-                "KeyID": ("92ed150794387c03ce684574b1139a65"
-                          "94a34f895daaaf09fd8ea90a27cddb12"),
-                "Host": "foo.example.org",
-                "Enabled": True,
-                "Timeout": 300000,
-                "LastCheckedOK": "2019-02-03T00:00:00",
-                "Created": "2019-01-02T00:00:00",
-                "Interval": 120000,
-                "Fingerprint": ("778827225BA7DE539C5A"
-                                "7CFA59CFF7CDBD9A5920"),
-                "CheckerRunning": False,
-                "LastEnabled": "2019-01-03T00:00:00",
-                "ApprovalPending": False,
-                "ApprovedByDefault": True,
-                "LastApprovalRequest": "",
-                "ApprovalDelay": 0,
-                "ApprovalDuration": 1000,
-                "Checker": "fping -q -- %(host)s",
-                "ExtendedTimeout": 900000,
-                "Expires": "2019-02-04T00:00:00",
-                "LastCheckerStatus": 0,
-            },
-            "barbar": {
-                "Name": "barbar",
-                "KeyID": ("0558568eedd67d622f5c83b35a115f79"
-                          "6ab612cff5ad227247e46c2b020f441c"),
-                "Host": "192.0.2.3",
-                "Enabled": True,
-                "Timeout": 300000,
-                "LastCheckedOK": "2019-02-04T00:00:00",
-                "Created": "2019-01-03T00:00:00",
-                "Interval": 120000,
-                "Fingerprint": ("3E393AEAEFB84C7E89E2"
-                                "F547B3A107558FCA3A27"),
-                "CheckerRunning": True,
-                "LastEnabled": "2019-01-04T00:00:00",
-                "ApprovalPending": False,
-                "ApprovedByDefault": False,
-                "LastApprovalRequest": "2019-01-03T00:00:00",
-                "ApprovalDelay": 30000,
-                "ApprovalDuration": 1000,
-                "Checker": ":",
-                "ExtendedTimeout": 900000,
-                "Expires": "2019-02-05T00:00:00",
-                "LastCheckerStatus": -2,
-            },
-        }
-        return super(TestDumpJSONCmd, self).setUp()
-    def test_normal(self):
-        json_data = json.loads(DumpJSONCmd().output(None,
-                                                    self.clients))
-        self.assertDictEqual(json_data, self.expected_json)
-    def test_one_client(self):
-        clients = self.one_client
-        json_data = json.loads(DumpJSONCmd().output(None, clients))
-        expected_json = {"foo": self.expected_json["foo"]}
-        self.assertDictEqual(json_data, expected_json)
-
-class TestIsEnabledCmd(TestCmd):
-    def test_is_enabled(self):
-        self.assertTrue(all(IsEnabledCmd().is_enabled(client, properties)
-                            for client, properties in self.clients.items()))
-    def test_is_enabled_run_exits_successfully(self):
+        self.one_client = {"/clients/foo": self.client.attributes}
+
+    @property
+    def bus(self):
+        class Bus(object):
+            @staticmethod
+            def get_object(client_bus_name, path):
+                self.assertEqual(client_bus_name, dbus_busname)
+                return {
+                    # Note: "self" here is the TestCmd instance, not
+                    # the Bus instance, since this is a static method!
+                    "/clients/foo": self.client,
+                    "/clients/barbar": self.other_client,
+                }[path]
+        return Bus()
+
+
+class TestBaseCommands(TestCommand):
+
+    def test_IsEnabled(self):
+        self.assertTrue(all(command.IsEnabled().is_enabled(client,
+                                                      properties)
+                            for client, properties
+                            in self.clients.items()))
+
+    def test_IsEnabled_run_exits_successfully(self):
         with self.assertRaises(SystemExit) as e:
-            IsEnabledCmd().run(None, None, self.one_client)
+            command.IsEnabled().run(self.one_client)
         if e.exception.code is not None:
             self.assertEqual(e.exception.code, 0)
         else:
             self.assertIsNone(e.exception.code)
-    def test_is_enabled_run_exits_with_failure(self):
+
+    def test_IsEnabled_run_exits_with_failure(self):
         self.client.attributes["Enabled"] = dbus.Boolean(False)
         with self.assertRaises(SystemExit) as e:
-            IsEnabledCmd().run(None, None, self.one_client)
+            command.IsEnabled().run(self.one_client)
         if isinstance(e.exception.code, int):
             self.assertNotEqual(e.exception.code, 0)
         else:
             self.assertIsNotNone(e.exception.code)
 
-class TestRemoveCmd(TestCmd):
-    def test_remove(self):
+    def test_Approve(self):
+        command.Approve().run(self.clients, self.bus)
+        for clientpath in self.clients:
+            client = self.bus.get_object(dbus_busname, clientpath)
+            self.assertIn(("Approve", (True, client_dbus_interface)),
+                          client.calls)
+
+    def test_Deny(self):
+        command.Deny().run(self.clients, self.bus)
+        for clientpath in self.clients:
+            client = self.bus.get_object(dbus_busname, clientpath)
+            self.assertIn(("Approve", (False, client_dbus_interface)),
+                          client.calls)
+
+    def test_Remove(self):
         class MockMandos(object):
             def __init__(self):
                 self.calls = []
             def RemoveClient(self, dbus_path):
                 self.calls.append(("RemoveClient", (dbus_path,)))
         mandos = MockMandos()
-        super(TestRemoveCmd, self).setUp()
-        RemoveCmd().run(mandos, None, self.clients)
-        self.assertEqual(len(mandos.calls), 2)
-        for client in self.clients:
-            self.assertIn(("RemoveClient",
-                           (client.__dbus_object_path__,)),
+        command.Remove().run(self.clients, self.bus, mandos)
+        for clientpath in self.clients:
+            self.assertIn(("RemoveClient", (clientpath,)),
                           mandos.calls)
 
-class TestApproveCmd(TestCmd):
-    def test_approve(self):
-        ApproveCmd().run(None, None, self.clients)
-        for client in self.clients:
-            self.assertIn(("Approve", (True, client_interface)),
-                          client.calls)
-
-class TestDenyCmd(TestCmd):
-    def test_deny(self):
-        DenyCmd().run(None, None, self.clients)
-        for client in self.clients:
-            self.assertIn(("Approve", (False, client_interface)),
-                          client.calls)
-
-class TestEnableCmd(TestCmd):
-    def test_enable(self):
-        for client in self.clients:
-            client.attributes["Enabled"] = False
-
-        EnableCmd().run(None, None, self.clients)
-
-        for client in self.clients:
-            self.assertTrue(client.attributes["Enabled"])
-
-class TestDisableCmd(TestCmd):
-    def test_disable(self):
-        DisableCmd().run(None, None, self.clients)
-
-        for client in self.clients:
-            self.assertFalse(client.attributes["Enabled"])
-
-class Unique(object):
-    """Class for objects which exist only to be unique objects, since
-unittest.mock.sentinel only exists in Python 3.3"""
-
-class TestPropertyCmd(TestCmd):
-    """Abstract class for tests of PropertyCmd classes"""
+    expected_json = {
+        "foo": {
+            "Name": "foo",
+            "KeyID": ("92ed150794387c03ce684574b1139a65"
+                      "94a34f895daaaf09fd8ea90a27cddb12"),
+            "Host": "foo.example.org",
+            "Enabled": True,
+            "Timeout": 300000,
+            "LastCheckedOK": "2019-02-03T00:00:00",
+            "Created": "2019-01-02T00:00:00",
+            "Interval": 120000,
+            "Fingerprint": ("778827225BA7DE539C5A"
+                            "7CFA59CFF7CDBD9A5920"),
+            "CheckerRunning": False,
+            "LastEnabled": "2019-01-03T00:00:00",
+            "ApprovalPending": False,
+            "ApprovedByDefault": True,
+            "LastApprovalRequest": "",
+            "ApprovalDelay": 0,
+            "ApprovalDuration": 1000,
+            "Checker": "fping -q -- %(host)s",
+            "ExtendedTimeout": 900000,
+            "Expires": "2019-02-04T00:00:00",
+            "LastCheckerStatus": 0,
+        },
+        "barbar": {
+            "Name": "barbar",
+            "KeyID": ("0558568eedd67d622f5c83b35a115f79"
+                      "6ab612cff5ad227247e46c2b020f441c"),
+            "Host": "192.0.2.3",
+            "Enabled": True,
+            "Timeout": 300000,
+            "LastCheckedOK": "2019-02-04T00:00:00",
+            "Created": "2019-01-03T00:00:00",
+            "Interval": 120000,
+            "Fingerprint": ("3E393AEAEFB84C7E89E2"
+                            "F547B3A107558FCA3A27"),
+            "CheckerRunning": True,
+            "LastEnabled": "2019-01-04T00:00:00",
+            "ApprovalPending": False,
+            "ApprovedByDefault": False,
+            "LastApprovalRequest": "2019-01-03T00:00:00",
+            "ApprovalDelay": 30000,
+            "ApprovalDuration": 93785000,
+            "Checker": ":",
+            "ExtendedTimeout": 900000,
+            "Expires": "2019-02-05T00:00:00",
+            "LastCheckerStatus": -2,
+        },
+    }
+
+    def test_DumpJSON_normal(self):
+        output = command.DumpJSON().output(self.clients.values())
+        json_data = json.loads(output)
+        self.assertDictEqual(json_data, self.expected_json)
+
+    def test_DumpJSON_one_client(self):
+        output = command.DumpJSON().output(self.one_client.values())
+        json_data = json.loads(output)
+        expected_json = {"foo": self.expected_json["foo"]}
+        self.assertDictEqual(json_data, expected_json)
+
+    def test_PrintTable_normal(self):
+        output = command.PrintTable().output(self.clients.values())
+        expected_output = "\n".join((
+            "Name   Enabled Timeout  Last Successful Check",
+            "foo    Yes     00:05:00 2019-02-03T00:00:00  ",
+            "barbar Yes     00:05:00 2019-02-04T00:00:00  ",
+        ))
+        self.assertEqual(output, expected_output)
+
+    def test_PrintTable_verbose(self):
+        output = command.PrintTable(verbose=True).output(
+            self.clients.values())
+        columns = (
+            (
+                "Name   ",
+                "foo    ",
+                "barbar ",
+            ),(
+                "Enabled ",
+                "Yes     ",
+                "Yes     ",
+            ),(
+                "Timeout  ",
+                "00:05:00 ",
+                "00:05:00 ",
+            ),(
+                "Last Successful Check ",
+                "2019-02-03T00:00:00   ",
+                "2019-02-04T00:00:00   ",
+            ),(
+                "Created             ",
+                "2019-01-02T00:00:00 ",
+                "2019-01-03T00:00:00 ",
+            ),(
+                "Interval ",
+                "00:02:00 ",
+                "00:02:00 ",
+            ),(
+                "Host            ",
+                "foo.example.org ",
+                "192.0.2.3       ",
+            ),(
+                ("Key ID                                             "
+                 "              "),
+                ("92ed150794387c03ce684574b1139a6594a34f895daaaf09fd8"
+                 "ea90a27cddb12 "),
+                ("0558568eedd67d622f5c83b35a115f796ab612cff5ad227247e"
+                 "46c2b020f441c "),
+            ),(
+                "Fingerprint                              ",
+                "778827225BA7DE539C5A7CFA59CFF7CDBD9A5920 ",
+                "3E393AEAEFB84C7E89E2F547B3A107558FCA3A27 ",
+            ),(
+                "Check Is Running ",
+                "No               ",
+                "Yes              ",
+            ),(
+                "Last Enabled        ",
+                "2019-01-03T00:00:00 ",
+                "2019-01-04T00:00:00 ",
+            ),(
+                "Approval Is Pending ",
+                "No                  ",
+                "No                  ",
+            ),(
+                "Approved By Default ",
+                "Yes                 ",
+                "No                  ",
+            ),(
+                "Last Approval Request ",
+                "                      ",
+                "2019-01-03T00:00:00   ",
+            ),(
+                "Approval Delay ",
+                "00:00:00       ",
+                "00:00:30       ",
+            ),(
+                "Approval Duration ",
+                "00:00:01          ",
+                "1T02:03:05        ",
+            ),(
+                "Checker              ",
+                "fping -q -- %(host)s ",
+                ":                    ",
+            ),(
+                "Extended Timeout ",
+                "00:15:00         ",
+                "00:15:00         ",
+            ),(
+                "Expires             ",
+                "2019-02-04T00:00:00 ",
+                "2019-02-05T00:00:00 ",
+            ),(
+                "Last Checker Status",
+                "0                  ",
+                "-2                 ",
+            )
+        )
+        num_lines = max(len(rows) for rows in columns)
+        expected_output = "\n".join("".join(rows[line]
+                                            for rows in columns)
+                                    for line in range(num_lines))
+        self.assertEqual(output, expected_output)
+
+    def test_PrintTable_one_client(self):
+        output = command.PrintTable().output(self.one_client.values())
+        expected_output = "\n".join((
+            "Name Enabled Timeout  Last Successful Check",
+            "foo  Yes     00:05:00 2019-02-03T00:00:00  ",
+        ))
+        self.assertEqual(output, expected_output)
+
+
+class TestPropertyCmd(TestCommand):
+    """Abstract class for tests of command.Property classes"""
     def runTest(self):
         if not hasattr(self, "command"):
             return
@@ -1079,444 +1696,152 @@
                                 self.values_to_set)
         for value_to_set, value_to_get in zip(self.values_to_set,
                                               values_to_get):
-            for client in self.clients:
-                old_value = client.attributes[self.property]
-                self.assertNotIsInstance(old_value, Unique)
-                client.attributes[self.property] = Unique()
+            for clientpath in self.clients:
+                client = self.bus.get_object(dbus_busname, clientpath)
+                old_value = client.attributes[self.propname]
+                client.attributes[self.propname] = self.Unique()
             self.run_command(value_to_set, self.clients)
-            for client in self.clients:
-                value = client.attributes[self.property]
-                self.assertNotIsInstance(value, Unique)
+            for clientpath in self.clients:
+                client = self.bus.get_object(dbus_busname, clientpath)
+                value = client.attributes[self.propname]
+                self.assertNotIsInstance(value, self.Unique)
                 self.assertEqual(value, value_to_get)
+
+    class Unique(object):
+        """Class for objects which exist only to be unique objects,
+since unittest.mock.sentinel only exists in Python 3.3"""
+
     def run_command(self, value, clients):
-        self.command().run(None, None, clients)
+        self.command().run(clients, self.bus)
+
+
+class TestEnableCmd(TestPropertyCmd):
+    command = command.Enable
+    propname = "Enabled"
+    values_to_set = [dbus.Boolean(True)]
+
+
+class TestDisableCmd(TestPropertyCmd):
+    command = command.Disable
+    propname = "Enabled"
+    values_to_set = [dbus.Boolean(False)]
+
 
 class TestBumpTimeoutCmd(TestPropertyCmd):
-    command = BumpTimeoutCmd
-    property = "LastCheckedOK"
+    command = command.BumpTimeout
+    propname = "LastCheckedOK"
     values_to_set = [""]
 
+
 class TestStartCheckerCmd(TestPropertyCmd):
-    command = StartCheckerCmd
-    property = "CheckerRunning"
+    command = command.StartChecker
+    propname = "CheckerRunning"
     values_to_set = [dbus.Boolean(True)]
 
+
 class TestStopCheckerCmd(TestPropertyCmd):
-    command = StopCheckerCmd
-    property = "CheckerRunning"
+    command = command.StopChecker
+    propname = "CheckerRunning"
     values_to_set = [dbus.Boolean(False)]
 
+
 class TestApproveByDefaultCmd(TestPropertyCmd):
-    command = ApproveByDefaultCmd
-    property = "ApprovedByDefault"
+    command = command.ApproveByDefault
+    propname = "ApprovedByDefault"
     values_to_set = [dbus.Boolean(True)]
 
+
 class TestDenyByDefaultCmd(TestPropertyCmd):
-    command = DenyByDefaultCmd
-    property = "ApprovedByDefault"
+    command = command.DenyByDefault
+    propname = "ApprovedByDefault"
     values_to_set = [dbus.Boolean(False)]
 
-class TestValueArgumentPropertyCmd(TestPropertyCmd):
-    """Abstract class for tests of PropertyCmd classes using the
-ValueArgumentMixIn"""
+
+class TestPropertyValueCmd(TestPropertyCmd):
+    """Abstract class for tests of PropertyValueCmd classes"""
+
     def runTest(self):
-        if type(self) is TestValueArgumentPropertyCmd:
+        if type(self) is TestPropertyValueCmd:
             return
-        return super(TestValueArgumentPropertyCmd, self).runTest()
+        return super(TestPropertyValueCmd, self).runTest()
+
     def run_command(self, value, clients):
-        self.command(value).run(None, None, clients)
-
-class TestSetCheckerCmd(TestValueArgumentPropertyCmd):
-    command = SetCheckerCmd
-    property = "Checker"
+        self.command(value).run(clients, self.bus)
+
+
+class TestSetCheckerCmd(TestPropertyValueCmd):
+    command = command.SetChecker
+    propname = "Checker"
     values_to_set = ["", ":", "fping -q -- %s"]
 
-class TestSetHostCmd(TestValueArgumentPropertyCmd):
-    command = SetHostCmd
-    property = "Host"
+
+class TestSetHostCmd(TestPropertyValueCmd):
+    command = command.SetHost
+    propname = "Host"
     values_to_set = ["192.0.2.3", "foo.example.org"]
 
-class TestSetSecretCmd(TestValueArgumentPropertyCmd):
-    command = SetSecretCmd
-    property = "Secret"
+
+class TestSetSecretCmd(TestPropertyValueCmd):
+    command = command.SetSecret
+    propname = "Secret"
     values_to_set = [io.BytesIO(b""),
                      io.BytesIO(b"secret\0xyzzy\nbar")]
     values_to_get = [b"", b"secret\0xyzzy\nbar"]
 
-class TestSetTimeoutCmd(TestValueArgumentPropertyCmd):
-    command = SetTimeoutCmd
-    property = "Timeout"
-    values_to_set = [datetime.timedelta(),
-                     datetime.timedelta(minutes=5),
-                     datetime.timedelta(seconds=1),
-                     datetime.timedelta(weeks=1),
-                     datetime.timedelta(weeks=52)]
-    values_to_get = [0, 300000, 1000, 604800000, 31449600000]
-
-class TestSetExtendedTimeoutCmd(TestValueArgumentPropertyCmd):
-    command = SetExtendedTimeoutCmd
-    property = "ExtendedTimeout"
-    values_to_set = [datetime.timedelta(),
-                     datetime.timedelta(minutes=5),
-                     datetime.timedelta(seconds=1),
-                     datetime.timedelta(weeks=1),
-                     datetime.timedelta(weeks=52)]
-    values_to_get = [0, 300000, 1000, 604800000, 31449600000]
-
-class TestSetIntervalCmd(TestValueArgumentPropertyCmd):
-    command = SetIntervalCmd
-    property = "Interval"
-    values_to_set = [datetime.timedelta(),
-                     datetime.timedelta(minutes=5),
-                     datetime.timedelta(seconds=1),
-                     datetime.timedelta(weeks=1),
-                     datetime.timedelta(weeks=52)]
-    values_to_get = [0, 300000, 1000, 604800000, 31449600000]
-
-class TestSetApprovalDelayCmd(TestValueArgumentPropertyCmd):
-    command = SetApprovalDelayCmd
-    property = "ApprovalDelay"
-    values_to_set = [datetime.timedelta(),
-                     datetime.timedelta(minutes=5),
-                     datetime.timedelta(seconds=1),
-                     datetime.timedelta(weeks=1),
-                     datetime.timedelta(weeks=52)]
-    values_to_get = [0, 300000, 1000, 604800000, 31449600000]
-
-class TestSetApprovalDurationCmd(TestValueArgumentPropertyCmd):
-    command = SetApprovalDurationCmd
-    property = "ApprovalDuration"
-    values_to_set = [datetime.timedelta(),
-                     datetime.timedelta(minutes=5),
-                     datetime.timedelta(seconds=1),
-                     datetime.timedelta(weeks=1),
-                     datetime.timedelta(weeks=52)]
-    values_to_get = [0, 300000, 1000, 604800000, 31449600000]
-
-class Test_command_from_options(unittest.TestCase):
-    def setUp(self):
-        self.parser = argparse.ArgumentParser()
-        add_command_line_options(self.parser)
-    def assert_command_from_args(self, args, command_cls, **cmd_attrs):
-        """Assert that parsing ARGS should result in an instance of
-COMMAND_CLS with (optionally) all supplied attributes (CMD_ATTRS)."""
-        options = self.parser.parse_args(args)
-        check_option_syntax(self.parser, options)
-        commands = commands_from_options(options)
-        self.assertEqual(len(commands), 1)
-        command = commands[0]
-        self.assertIsInstance(command, command_cls)
-        for key, value in cmd_attrs.items():
-            self.assertEqual(getattr(command, key), value)
-    def test_print_table(self):
-        self.assert_command_from_args([], PrintTableCmd,
-                                      verbose=False)
-
-    def test_print_table_verbose(self):
-        self.assert_command_from_args(["--verbose"], PrintTableCmd,
-                                      verbose=True)
-
-    def test_print_table_verbose_short(self):
-        self.assert_command_from_args(["-v"], PrintTableCmd,
-                                      verbose=True)
-
-    def test_enable(self):
-        self.assert_command_from_args(["--enable", "foo"], EnableCmd)
-
-    def test_enable_short(self):
-        self.assert_command_from_args(["-e", "foo"], EnableCmd)
-
-    def test_disable(self):
-        self.assert_command_from_args(["--disable", "foo"],
-                                      DisableCmd)
-
-    def test_disable_short(self):
-        self.assert_command_from_args(["-d", "foo"], DisableCmd)
-
-    def test_bump_timeout(self):
-        self.assert_command_from_args(["--bump-timeout", "foo"],
-                                      BumpTimeoutCmd)
-
-    def test_bump_timeout_short(self):
-        self.assert_command_from_args(["-b", "foo"], BumpTimeoutCmd)
-
-    def test_start_checker(self):
-        self.assert_command_from_args(["--start-checker", "foo"],
-                                      StartCheckerCmd)
-
-    def test_stop_checker(self):
-        self.assert_command_from_args(["--stop-checker", "foo"],
-                                      StopCheckerCmd)
-
-    def test_remove(self):
-        self.assert_command_from_args(["--remove", "foo"],
-                                      RemoveCmd)
-
-    def test_remove_short(self):
-        self.assert_command_from_args(["-r", "foo"], RemoveCmd)
-
-    def test_checker(self):
-        self.assert_command_from_args(["--checker", ":", "foo"],
-                                      SetCheckerCmd, value_to_set=":")
-
-    def test_checker_empty(self):
-        self.assert_command_from_args(["--checker", "", "foo"],
-                                      SetCheckerCmd, value_to_set="")
-
-    def test_checker_short(self):
-        self.assert_command_from_args(["-c", ":", "foo"],
-                                      SetCheckerCmd, value_to_set=":")
-
-    def test_timeout(self):
-        self.assert_command_from_args(["--timeout", "PT5M", "foo"],
-                                      SetTimeoutCmd,
-                                      value_to_set=300000)
-
-    def test_timeout_short(self):
-        self.assert_command_from_args(["-t", "PT5M", "foo"],
-                                      SetTimeoutCmd,
-                                      value_to_set=300000)
-
-    def test_extended_timeout(self):
-        self.assert_command_from_args(["--extended-timeout", "PT15M",
-                                       "foo"],
-                                      SetExtendedTimeoutCmd,
-                                      value_to_set=900000)
-
-    def test_interval(self):
-        self.assert_command_from_args(["--interval", "PT2M", "foo"],
-                                      SetIntervalCmd,
-                                      value_to_set=120000)
-
-    def test_interval_short(self):
-        self.assert_command_from_args(["-i", "PT2M", "foo"],
-                                      SetIntervalCmd,
-                                      value_to_set=120000)
-
-    def test_approve_by_default(self):
-        self.assert_command_from_args(["--approve-by-default", "foo"],
-                                      ApproveByDefaultCmd)
-
-    def test_deny_by_default(self):
-        self.assert_command_from_args(["--deny-by-default", "foo"],
-                                      DenyByDefaultCmd)
-
-    def test_approval_delay(self):
-        self.assert_command_from_args(["--approval-delay", "PT30S",
-                                       "foo"], SetApprovalDelayCmd,
-                                      value_to_set=30000)
-
-    def test_approval_duration(self):
-        self.assert_command_from_args(["--approval-duration", "PT1S",
-                                       "foo"], SetApprovalDurationCmd,
-                                      value_to_set=1000)
-
-    def test_host(self):
-        self.assert_command_from_args(["--host", "foo.example.org",
-                                       "foo"], SetHostCmd,
-                                      value_to_set="foo.example.org")
-
-    def test_host_short(self):
-        self.assert_command_from_args(["-H", "foo.example.org",
-                                       "foo"], SetHostCmd,
-                                      value_to_set="foo.example.org")
-
-    def test_secret_devnull(self):
-        self.assert_command_from_args(["--secret", os.path.devnull,
-                                       "foo"], SetSecretCmd,
-                                      value_to_set=b"")
-
-    def test_secret_tempfile(self):
-        with tempfile.NamedTemporaryFile(mode="r+b") as f:
-            value = b"secret\0xyzzy\nbar"
-            f.write(value)
-            f.seek(0)
-            self.assert_command_from_args(["--secret", f.name,
-                                           "foo"], SetSecretCmd,
-                                          value_to_set=value)
-
-    def test_secret_devnull_short(self):
-        self.assert_command_from_args(["-s", os.path.devnull, "foo"],
-                                      SetSecretCmd, value_to_set=b"")
-
-    def test_secret_tempfile_short(self):
-        with tempfile.NamedTemporaryFile(mode="r+b") as f:
-            value = b"secret\0xyzzy\nbar"
-            f.write(value)
-            f.seek(0)
-            self.assert_command_from_args(["-s", f.name, "foo"],
-                                          SetSecretCmd,
-                                          value_to_set=value)
-
-    def test_approve(self):
-        self.assert_command_from_args(["--approve", "foo"],
-                                      ApproveCmd)
-
-    def test_approve_short(self):
-        self.assert_command_from_args(["-A", "foo"], ApproveCmd)
-
-    def test_deny(self):
-        self.assert_command_from_args(["--deny", "foo"], DenyCmd)
-
-    def test_deny_short(self):
-        self.assert_command_from_args(["-D", "foo"], DenyCmd)
-
-    def test_dump_json(self):
-        self.assert_command_from_args(["--dump-json"], DumpJSONCmd)
-
-    def test_is_enabled(self):
-        self.assert_command_from_args(["--is-enabled", "foo"],
-                                      IsEnabledCmd)
-
-    def test_is_enabled_short(self):
-        self.assert_command_from_args(["-V", "foo"], IsEnabledCmd)
-
-    def test_deny_before_remove(self):
-        options = self.parser.parse_args(["--deny", "--remove", "foo"])
-        check_option_syntax(self.parser, options)
-        commands = commands_from_options(options)
-        self.assertEqual(len(commands), 2)
-        self.assertIsInstance(commands[0], DenyCmd)
-        self.assertIsInstance(commands[1], RemoveCmd)
-
-    def test_deny_before_remove_reversed(self):
-        options = self.parser.parse_args(["--remove", "--deny", "--all"])
-        check_option_syntax(self.parser, options)
-        commands = commands_from_options(options)
-        self.assertEqual(len(commands), 2)
-        self.assertIsInstance(commands[0], DenyCmd)
-        self.assertIsInstance(commands[1], RemoveCmd)
-
-
-class Test_check_option_syntax(unittest.TestCase):
-    # This mostly corresponds to the definition from has_actions() in
-    # check_option_syntax()
-    actions = {
-        # The actual values set here are not that important, but we do
-        # at least stick to the correct types, even though they are
-        # never used
-        "enable": True,
-        "disable": True,
-        "bump_timeout": True,
-        "start_checker": True,
-        "stop_checker": True,
-        "is_enabled": True,
-        "remove": True,
-        "checker": "x",
-        "timeout": datetime.timedelta(),
-        "extended_timeout": datetime.timedelta(),
-        "interval": datetime.timedelta(),
-        "approved_by_default": True,
-        "approval_delay": datetime.timedelta(),
-        "approval_duration": datetime.timedelta(),
-        "host": "x",
-        "secret": io.BytesIO(b"x"),
-        "approve": True,
-        "deny": True,
-    }
-
-    def setUp(self):
-        self.parser = argparse.ArgumentParser()
-        add_command_line_options(self.parser)
-
-    @contextlib.contextmanager
-    def assertParseError(self):
-        with self.assertRaises(SystemExit) as e:
-            with self.temporarily_suppress_stderr():
-                yield
-        # Exit code from argparse is guaranteed to be "2".  Reference:
-        # https://docs.python.org/3/library/argparse.html#exiting-methods
-        self.assertEqual(e.exception.code, 2)
-
-    @staticmethod
-    @contextlib.contextmanager
-    def temporarily_suppress_stderr():
-        null = os.open(os.path.devnull, os.O_RDWR)
-        stderrcopy = os.dup(sys.stderr.fileno())
-        os.dup2(null, sys.stderr.fileno())
-        os.close(null)
-        try:
-            yield
-        finally:
-            # restore stderr
-            os.dup2(stderrcopy, sys.stderr.fileno())
-            os.close(stderrcopy)
-
-    def check_option_syntax(self, options):
-        check_option_syntax(self.parser, options)
-
-    def test_actions_requires_client_or_all(self):
-        for action, value in self.actions.items():
-            options = self.parser.parse_args()
-            setattr(options, action, value)
-            with self.assertParseError():
-                self.check_option_syntax(options)
-
-    def test_actions_conflicts_with_verbose(self):
-        for action, value in self.actions.items():
-            options = self.parser.parse_args()
-            setattr(options, action, value)
-            options.verbose = True
-            with self.assertParseError():
-                self.check_option_syntax(options)
-
-    def test_dump_json_conflicts_with_verbose(self):
-        options = self.parser.parse_args()
-        options.dump_json = True
-        options.verbose = True
-        with self.assertParseError():
-            self.check_option_syntax(options)
-
-    def test_dump_json_conflicts_with_action(self):
-        for action, value in self.actions.items():
-            options = self.parser.parse_args()
-            setattr(options, action, value)
-            options.dump_json = True
-            with self.assertParseError():
-                self.check_option_syntax(options)
-
-    def test_all_can_not_be_alone(self):
-        options = self.parser.parse_args()
-        options.all = True
-        with self.assertParseError():
-            self.check_option_syntax(options)
-
-    def test_all_is_ok_with_any_action(self):
-        for action, value in self.actions.items():
-            options = self.parser.parse_args()
-            setattr(options, action, value)
-            options.all = True
-            self.check_option_syntax(options)
-
-    def test_is_enabled_fails_without_client(self):
-        options = self.parser.parse_args()
-        options.is_enabled = True
-        with self.assertParseError():
-            self.check_option_syntax(options)
-
-    def test_is_enabled_works_with_one_client(self):
-        options = self.parser.parse_args()
-        options.is_enabled = True
-        options.client = ["foo"]
-        self.check_option_syntax(options)
-
-    def test_is_enabled_fails_with_two_clients(self):
-        options = self.parser.parse_args()
-        options.is_enabled = True
-        options.client = ["foo", "barbar"]
-        with self.assertParseError():
-            self.check_option_syntax(options)
-
-    def test_remove_can_only_be_combined_with_action_deny(self):
-        for action, value in self.actions.items():
-            if action in {"remove", "deny"}:
-                continue
-            options = self.parser.parse_args()
-            setattr(options, action, value)
-            options.all = True
-            options.remove = True
-            with self.assertParseError():
-                self.check_option_syntax(options)
+
+class TestSetTimeoutCmd(TestPropertyValueCmd):
+    command = command.SetTimeout
+    propname = "Timeout"
+    values_to_set = [datetime.timedelta(),
+                     datetime.timedelta(minutes=5),
+                     datetime.timedelta(seconds=1),
+                     datetime.timedelta(weeks=1),
+                     datetime.timedelta(weeks=52)]
+    values_to_get = [0, 300000, 1000, 604800000, 31449600000]
+
+
+class TestSetExtendedTimeoutCmd(TestPropertyValueCmd):
+    command = command.SetExtendedTimeout
+    propname = "ExtendedTimeout"
+    values_to_set = [datetime.timedelta(),
+                     datetime.timedelta(minutes=5),
+                     datetime.timedelta(seconds=1),
+                     datetime.timedelta(weeks=1),
+                     datetime.timedelta(weeks=52)]
+    values_to_get = [0, 300000, 1000, 604800000, 31449600000]
+
+
+class TestSetIntervalCmd(TestPropertyValueCmd):
+    command = command.SetInterval
+    propname = "Interval"
+    values_to_set = [datetime.timedelta(),
+                     datetime.timedelta(minutes=5),
+                     datetime.timedelta(seconds=1),
+                     datetime.timedelta(weeks=1),
+                     datetime.timedelta(weeks=52)]
+    values_to_get = [0, 300000, 1000, 604800000, 31449600000]
+
+
+class TestSetApprovalDelayCmd(TestPropertyValueCmd):
+    command = command.SetApprovalDelay
+    propname = "ApprovalDelay"
+    values_to_set = [datetime.timedelta(),
+                     datetime.timedelta(minutes=5),
+                     datetime.timedelta(seconds=1),
+                     datetime.timedelta(weeks=1),
+                     datetime.timedelta(weeks=52)]
+    values_to_get = [0, 300000, 1000, 604800000, 31449600000]
+
+
+class TestSetApprovalDurationCmd(TestPropertyValueCmd):
+    command = command.SetApprovalDuration
+    propname = "ApprovalDuration"
+    values_to_set = [datetime.timedelta(),
+                     datetime.timedelta(minutes=5),
+                     datetime.timedelta(seconds=1),
+                     datetime.timedelta(weeks=1),
+                     datetime.timedelta(weeks=52)]
+    values_to_get = [0, 300000, 1000, 604800000, 31449600000]
 
 
 
@@ -1537,8 +1862,11 @@
     return tests
 
 if __name__ == "__main__":
-    if should_only_run_tests():
-        # Call using ./tdd-python-script --check [--verbose]
-        unittest.main()
-    else:
-        main()
+    try:
+        if should_only_run_tests():
+            # Call using ./tdd-python-script --check [--verbose]
+            unittest.main()
+        else:
+            main()
+    finally:
+        logging.shutdown()

